
__copyright__ = 'Copyright 2013-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

from typing import Optional, List

import threading     as mt
import radical.utils as ru

from .constants  import BUSY, DOWN
from .utils.misc import FastTypedDict


LABEL                  = 'label'
DESCRIPTION            = 'description'
NOTES                  = 'notes'
DEFAULT_SCHEMA         = 'default_schema'
SCHEMAS                = 'schemas'
JOB_MANAGER_ENDPOINT   = 'job_manager_endpoint'
JOB_MANAGER_HOP        = 'job_manager_hop'
FILESYSTEM_ENDPOINT    = 'filesystem_endpoint'
DEFAULT_REMOTE_WORKDIR = 'default_remote_workdir'
DEFAULT_QUEUE          = 'default_queue'
RESOURCE_MANAGER       = 'resource_manager'
AGENT_CONFIG           = 'agent_config'
AGENT_SCHEDULER        = 'agent_scheduler'
AGENT_SPAWNER          = 'agent_spawner'
PRE_BOOTSTRAP_0        = 'pre_bootstrap_0'
PRE_BOOTSTRAP_1        = 'pre_bootstrap_1'
RP_VERSION             = 'rp_version'
VIRTENV                = 'virtenv'
VIRTENV_MODE           = 'virtenv_mode'
PYTHON_DIST            = 'python_dist'
PYTHON_INTERPRETER     = 'python_interpreter'
LAUNCH_METHODS         = 'launch_methods'
LFS_PATH_PER_NODE      = 'lfs_path_per_node'
LFS_SIZE_PER_NODE      = 'lfs_size_per_node'
TASK_TMP               = 'task_tmp'
MEM_PER_NODE           = 'mem_per_node'
CORES_PER_NODE         = 'cores_per_node'
GPUS_PER_NODE          = 'gpus_per_node'
NUMA_DOMAIN_MAP        = 'numa_domain_map'
SYSTEM_ARCHITECTURE    = 'system_architecture'
SCATTERED              = 'scattered'

FAKE_RESOURCES         = 'fake_resources'
MANDATORY_ARGS         = 'mandatory_args'
FORWARD_TUNNEL_ENDPOINT = 'forward_tunnel_endpoint'

NEW_SESSION_PER_TASK   = 'new_session_per_task'
TASK_PRE_LAUNCH        = 'task_pre_launch'
TASK_POST_LAUNCH       = 'task_post_launch'
TASK_PRE_EXEC          = 'task_pre_exec'
TASK_POST_EXEC         = 'task_post_exec'

RAPTOR                 = 'raptor'
RAPTOR_HB_DELAY        = 'hb_delay'
RAPTOR_HB_TIMEOUT      = 'hb_timeout'
RAPTOR_HB_FREQUENCY    = 'hb_frequency'

ENDPOINTS_DEFAULT      = {JOB_MANAGER_ENDPOINT: 'fork://localhost/',
                          FILESYSTEM_ENDPOINT : 'file://localhost/'}


# ------------------------------------------------------------------------------
#
class RaptorConfig(FastTypedDict):

    _schema = {
        RAPTOR_HB_DELAY    : int,
        RAPTOR_HB_TIMEOUT  : int,
        RAPTOR_HB_FREQUENCY: int,
    }

    _defaults = {
        RAPTOR_HB_DELAY    : 5,
        RAPTOR_HB_TIMEOUT  : 500,
        RAPTOR_HB_FREQUENCY: 1000,
    }


# ------------------------------------------------------------------------------
#
class AccessSchema(FastTypedDict):

    _schema = {
        JOB_MANAGER_ENDPOINT: str,
        JOB_MANAGER_HOP     : str,
        FILESYSTEM_ENDPOINT : str,
    }

    _defaults = {
        JOB_MANAGER_ENDPOINT: None,
        JOB_MANAGER_HOP     : None,
        FILESYSTEM_ENDPOINT : None,
    }


# ------------------------------------------------------------------------------
#
class ResourceConfig(FastTypedDict):
    '''
    docstrings goes here
    '''

    _schema = {
        LABEL                  : str         ,
        DESCRIPTION            : str         ,
        NOTES                  : str         ,
        DEFAULT_SCHEMA         : str         ,
        SCHEMAS                : {str: AccessSchema},
        RAPTOR                 : RaptorConfig,

        # FIXME: AM - need to resolve since  in Session it is moved into RD
        #        `_get_resource_sandbox` ->  `KeyError: 'filesystem_endpoint'`
        JOB_MANAGER_ENDPOINT   : str         ,
        JOB_MANAGER_HOP        : str         ,
        FILESYSTEM_ENDPOINT    : str         ,

        DEFAULT_REMOTE_WORKDIR : str         ,
        DEFAULT_QUEUE          : str         ,
        RESOURCE_MANAGER       : str         ,
        AGENT_CONFIG           : str         ,
        AGENT_SCHEDULER        : str         ,
        AGENT_SPAWNER          : str         ,
        PRE_BOOTSTRAP_0        : [str]       ,
        PRE_BOOTSTRAP_1        : [str]       ,
        RP_VERSION             : str         ,
        VIRTENV                : str         ,
        VIRTENV_MODE           : str         ,
        PYTHON_DIST            : str         ,
        PYTHON_INTERPRETER     : str         ,
        LAUNCH_METHODS         : {str: None} ,
        LFS_PATH_PER_NODE      : str         ,
        LFS_SIZE_PER_NODE      : int         ,
        TASK_TMP               : str         ,
        MEM_PER_NODE           : int         ,
        CORES_PER_NODE         : int         ,
        GPUS_PER_NODE          : int         ,
        NUMA_DOMAIN_MAP        : {int: None} ,
        SYSTEM_ARCHITECTURE    : {str: None} ,
        SCATTERED              : bool        ,

        FAKE_RESOURCES         : bool        ,
        MANDATORY_ARGS         : [str]       ,
        FORWARD_TUNNEL_ENDPOINT: str         ,
        NEW_SESSION_PER_TASK   : bool        ,
        TASK_PRE_LAUNCH        : [str]       ,
        TASK_POST_LAUNCH       : [str]       ,
        TASK_PRE_EXEC          : [str]       ,
        TASK_POST_EXEC         : [str]       ,
    }

    _defaults = {
        LABEL                  : ''          ,
        DESCRIPTION            : ''          ,
        NOTES                  : ''          ,
        DEFAULT_SCHEMA         : ''          ,
        SCHEMAS                : list()      ,
        RAPTOR                 : RaptorConfig(),

        # FIXME: AM - need to resolve since in Session it is moved into RD
        #        `_get_resource_sandbox` -> `KeyError: 'filesystem_endpoint'`
        JOB_MANAGER_ENDPOINT   : None        ,
        JOB_MANAGER_HOP        : None        ,
        FILESYSTEM_ENDPOINT    : None        ,

        DEFAULT_REMOTE_WORKDIR : ''          ,
        DEFAULT_QUEUE          : ''          ,
        RESOURCE_MANAGER       : ''          ,
        AGENT_CONFIG           : 'default'   ,
        AGENT_SCHEDULER        : 'CONTINUOUS',
        AGENT_SPAWNER          : 'POPEN'     ,
        PRE_BOOTSTRAP_0        : list()      ,
        PRE_BOOTSTRAP_1        : list()      ,
        RP_VERSION             : 'installed' ,
        VIRTENV                : ''          ,
        VIRTENV_MODE           : 'local'     ,
        PYTHON_DIST            : 'default'   ,
        PYTHON_INTERPRETER     : ''          ,
        LAUNCH_METHODS         : dict()      ,
        LFS_PATH_PER_NODE      : ''          ,
        LFS_SIZE_PER_NODE      : 0           ,
        TASK_TMP               : ''          ,
        MEM_PER_NODE           : 0           ,
        CORES_PER_NODE         : 0           ,
        GPUS_PER_NODE          : 0           ,
        NUMA_DOMAIN_MAP        : dict()      ,
        SYSTEM_ARCHITECTURE    : dict()      ,
        SCATTERED              : True        ,

        FAKE_RESOURCES         : False       ,
        MANDATORY_ARGS         : list()      ,
        FORWARD_TUNNEL_ENDPOINT: ''          ,
        NEW_SESSION_PER_TASK   : True        ,
        TASK_PRE_LAUNCH        : list()      ,
        TASK_POST_LAUNCH       : list()      ,
        TASK_PRE_EXEC          : list()      ,
        TASK_POST_EXEC         : list()      ,
    }


# ------------------------------------------------------------------------------
#
class RO(FastTypedDict):

    INDEX      = 'index'
    OCCUPATION = 'occupation'

    _schema = {
        INDEX     : int,
        OCCUPATION: float,
    }

    _defaults = {
        INDEX     : 0,
        OCCUPATION: None,
    }

    def __init__(self, from_dict: Optional[dict] = None,
                       **kwargs) -> None:

        super().__init__(from_dict=from_dict, **kwargs)


    def __str__(self):
        if self.occupation == DOWN:
            return '%d:----' % self.index
        return '%d:%.2f' % (self.index, self.occupation)

    def _verify(self):

        if self.occupation is None:
            raise ValueError('missing occupation: %s' % self.occupation)


ResourceOccupation = RO


# ------------------------------------------------------------------------------
#
class RankRequirements(FastTypedDict):

    N_CORES         = 'n_cores'
    CORE_OCCUPATION = 'core_occupation'
    N_GPUS          = 'n_gpus'
    GPU_OCCUPATION  = 'gpu_occupation'
    LFS             = 'lfs'
    MEM             = 'mem'
    NUMA            = 'numa'

    _schema = {
        N_CORES        : int,
        CORE_OCCUPATION: float,
        N_GPUS         : int,
        GPU_OCCUPATION : float,
        LFS            : int,
        MEM            : int,
        NUMA           : bool,
    }

    _defaults = {
        N_CORES        : 1,
        CORE_OCCUPATION: BUSY,
        N_GPUS         : 0,
        GPU_OCCUPATION : BUSY,
        LFS            : 0,
        MEM            : 0,
        NUMA           : False,
    }

    def __str__(self):
        return 'RR(%d:%.2f, %d:%.2f, %d, %d [%s])' % (
                self.n_cores, self.core_occupation,
                self.n_gpus,  self.gpu_occupation,
                self.lfs,     self.mem,
                'numa' if self.numa else 'non-numa')


    # comparison operators
    #
    # NOTE: we are not really interested in strict ordering here,
    #       but we do care about *fast* and approximate comparisons.
    #       'larger' in general means 'more or more specialized resources'.
    #
    # NOTE: NUMA setting is not considered for comparison
    #
    def __eq__(self, other: 'RankRequirements') -> bool:

        if  self.n_cores         == other.n_cores          and \
            self.n_gpus          == other.n_gpus           and \
            self.lfs             == other.lfs              and \
            self.mem             == other.mem              and \
            self.core_occupation == other.core_occupation  and \
            self.gpu_occupation  == other.gpu_occupation   :
            return True
        return False

    def __lt__(self, other: 'RankRequirements') -> bool:

        if  self.n_cores         <= other.n_cores          and \
            self.n_gpus          <= other.n_gpus           and \
            self.lfs             <= other.lfs              and \
            self.mem             <= other.mem              and \
            self.core_occupation <= other.core_occupation  and \
            self.gpu_occupation  <= other.gpu_occupation:
            return True
        return False

    def __le__(self, other: 'RankRequirements') -> bool:

        if  self.n_cores         <= other.n_cores          and \
            self.n_gpus          <= other.n_gpus           and \
            self.lfs             <= other.lfs              and \
            self.mem             <= other.mem              and \
            self.core_occupation <= other.core_occupation  and \
            self.gpu_occupation  <= other.gpu_occupation:
            return True
        return False

    def __gt__(self, other: 'RankRequirements') -> bool:

        if  self.n_cores         >= other.n_cores          and \
            self.n_gpus          >= other.n_gpus           and \
            self.lfs             >= other.lfs              and \
            self.mem             >= other.mem              and \
            self.core_occupation >= other.core_occupation  and \
            self.gpu_occupation  >= other.gpu_occupation:
            return True
        return False

    def __ge__(self, other: 'RankRequirements') -> bool:

        if  self.n_cores         >= other.n_cores          and \
            self.n_gpus          >= other.n_gpus           and \
            self.lfs             >= other.lfs              and \
            self.mem             >= other.mem              and \
            self.core_occupation >= other.core_occupation  and \
            self.gpu_occupation  >= other.gpu_occupation:
            return True
        return False


# ------------------------------------------------------------------------------
#
class Slot(FastTypedDict):

    CORES       = 'cores'
    GPUS        = 'gpus'
    LFS         = 'lfs'
    MEM         = 'mem'
    NODE_INDEX  = 'node_index'
    NODE_NAME   = 'node_name'
    VERSION     = 'version'  # use this to distinguish from old slot structure

    _schema = {
        CORES      : [RO],  # list of tuples [(core_id, core_occupation), ...]
        GPUS       : [RO],  # list of tuples [(gpu_id,  core_occupation), ...]
        LFS        : int,
        MEM        : int,
        NODE_INDEX : int,
        NODE_NAME  : str,
        VERSION    : int,
    }

    _defaults = {
        CORES      : list(),
        GPUS       : list(),
        LFS        : 0,
        MEM        : 0,
        NODE_INDEX : 0,
        NODE_NAME  : '',
        VERSION    : 1,
    }

    def __init__(self, from_dict: dict = None, **kwargs):

        if not from_dict:
            from_dict = kwargs

        if from_dict:

            cores = from_dict.get('cores')
            gpus  = from_dict.get('gpus')

            if cores:
                # this is much faster than `isinstance`
                if cores[0].__class__.__name__ == 'dict':
                    from_dict['cores'] =  [RO(d) for d in cores]

                elif isinstance(cores[0], int):
                    from_dict['cores'] =  [RO(index=i, occupation=BUSY)
                                                 for i in cores]

            if gpus:
                if gpus[0].__class__.__name__ == 'dict':
                    from_dict['gpus'] =  [RO(d) for d in gpus]

                elif isinstance(gpus[0], int):
                    from_dict['gpus'] =  [RO(index=i, occupation=BUSY)
                                                for i in gpus]


        super().__init__(from_dict, **kwargs)


# ------------------------------------------------------------------------------
#
class Node(FastTypedDict):
    '''
    Node resources as reported by the resource manager, used by the scheduler
    '''

    INDEX    = 'index'
    NAME     = 'name'
    CORES    = 'cores'
    GPUS     = 'gpus'
    LFS      = 'lfs'
    MEM      = 'mem'

    _schema = {
        INDEX    : int,
        NAME     : str,
        CORES    : [RO],
        GPUS     : [RO],
        LFS      : int,
        MEM      : int,
    }

    _defaults = {
        INDEX    : 0,
        NAME     : '',
        CORES    : list(),
        GPUS     : list(),
        LFS      : None,
        MEM      : None,
    }


    # --------------------------------------------------------------------------
    #
    def __init__(self, from_dict: dict):

        self.__lock__ = mt.RLock()

        cores = from_dict.get('cores')
        gpus  = from_dict.get('gpus')

        if cores:
            if not isinstance(cores[0], RO):
                from_dict['cores'] = [RO(index=i, occupation=o)
                                                    for i,o in enumerate(cores)]

        if gpus:
            if not isinstance(gpus[0], RO):
                from_dict['gpus'] = [RO(index=i, occupation=o)
                                                     for i,o in enumerate(gpus)]

        super().__init__(from_dict)


    # --------------------------------------------------------------------------
    #
    def _get_core_index(self, ro):

        for i, _ro in enumerate(self.cores):
            if _ro.index == ro.index:
                return i

        raise ValueError('invalid core index %s' % ro.index)


    # --------------------------------------------------------------------------
    #
    def _get_gpu_index(self, ro):

        for i, _ro in enumerate(self.gpus):
            if _ro.index == ro.index:
                return i

        raise ValueError('invalid gpu index %s' % ro.index)


    # --------------------------------------------------------------------------
    #
    def allocate_slot(self, slot  : Slot,
                            _check: bool = True) -> None:

        cores = slot.cores
        gpus  = slot.gpus
        lfs   = slot.lfs
        mem   = slot.mem

        if _check:
            assert self.index == slot.node_index
            assert self.name  == slot.node_name

        with self.__lock__:

            # we only need to perform consistency checks for slots which were not
            # created by `self.find_slot`
            if _check:

                # we allow for core indexes but convert into full occupancy then
                if cores and isinstance(cores[0], int):
                    cores = [RO(index=core, occupation=BUSY) for core in cores]

                if gpus and isinstance(gpus[0], int):
                    gpus = [RO(index=gpu, occupation=BUSY) for gpu in gpus]

                # make sure the selected cores exist, are not down, and
                # occupancy is compatible with the request
                for ro in cores:
                    assert ro.index < len(self.cores)
                    ro_available = BUSY - self.cores[ro.index].occupation
                    assert ro_available >= ro.occupation
                  # # DOWN check is covered by occupancy check
                  # assert self.cores[ro.index].occupation is not DOWN, \
                  #         'core %d is down' % ro.index

                for ro in gpus:
                    assert ro.index < len(self.gpus)
                    ro_available = BUSY - self.gpus[ro.index].occupation
                    assert ro_available >= ro.occupation
                  # # DOWN check is covered by occupancy check
                  # assert self.gpus[ro.index].occupation is not DOWN, \
                  #         'gpu %d is down' % ro.index

                if lfs: assert self.lfs >= lfs
                if mem: assert self.mem >= mem

            # slot is valid = apply the respective changes
            for ro in cores:
                c_idx = self._get_core_index(ro)
                self.cores[c_idx].occupation += ro.occupation

            for ro in gpus:
                g_idx = self._get_gpu_index(ro)
                self.gpus[g_idx].occupation += ro.occupation

            if self.lfs is not None: self.lfs -= lfs
            if self.mem is not None: self.mem -= mem


    # --------------------------------------------------------------------------
    #
    def deallocate_slot(self, slot : 'Slot') -> None:

        with self.__lock__:

            for ro in slot.cores:
                self.cores[ro.index].occupation -= ro.occupation
              # assert self.cores[ro.index].occupation >= 0.0, \
              #         'invalid core release: %s' % self

            for ro in slot.gpus:
                self.gpus[ro.index].occupation -= ro.occupation
              # assert self.gpus[ro.index].occupation >= 0.0, \
              #         'invalid gpu release: %s' % self

            self.lfs += slot.lfs
            self.mem += slot.mem


    # --------------------------------------------------------------------------
    #
    def find_slot(self, rr: RankRequirements) -> Optional[Slot]:

        with self.__lock__:

            cores = list()
            gpus  = list()

            # NOTE: the current mechanism will never use the same core or gpu
            #       multiple times for the created slot, even if the respective
            #       occupation would allow for it.
            if rr.n_cores:
                for ro in self.cores:
                    if ro.occupation is DOWN:
                        continue
                    if rr.core_occupation <= BUSY - ro.occupation:
                        cores.append(RO(index=ro.index,
                                        occupation=rr.core_occupation))
                    if len(cores) == rr.n_cores:
                        break

                if len(cores) < rr.n_cores:
                    return None

            if rr.n_gpus:
                for ro in self.gpus:
                    if ro.occupation is DOWN:
                        continue
                    if rr.gpu_occupation <= BUSY - ro.occupation:
                        gpus.append(RO(index=ro.index,
                                       occupation=rr.gpu_occupation))
                    if len(gpus) == rr.n_gpus:
                        break

                if len(gpus) < rr.n_gpus:
                    return None

            if self.lfs is not None:
                if rr.lfs and self.lfs < rr.lfs: return None

            if self.mem is not None:
                if rr.mem and self.mem < rr.mem: return None

            slot = Slot(cores=cores, gpus=gpus, lfs=rr.lfs, mem=rr.mem,
                        node_index=self.index, node_name=self.name)
            self.allocate_slot(slot, _check=False)

            return slot


# ------------------------------------------------------------------------------
#
class NodeList(FastTypedDict):

    NODES          = 'nodes'
    UNIFORM        = 'uniform'
    CORES_PER_NODE = 'cores_per_node'
    GPUS_PER_NODE  = 'gpus_per_node'
    LFS_PER_NODE   = 'lfs_per_node'
    MEM_PER_NODE   = 'mem_per_node'

    _schema = {
        NODES          : [Node],
        UNIFORM        : bool,
        CORES_PER_NODE : int,
        GPUS_PER_NODE  : int,
        LFS_PER_NODE   : int,
        MEM_PER_NODE   : int,

    }

    _defaults = {
        NODES          : list(),
        UNIFORM        : True,
        CORES_PER_NODE : None,
        GPUS_PER_NODE  : None,
        LFS_PER_NODE   : None,
        MEM_PER_NODE   : None,
    }


    # --------------------------------------------------------------------------
    #
    def __init__(self, from_dict: dict = None, **kwargs) -> None:

        super().__init__(from_dict=from_dict, **kwargs)

        self.__index__          = int()
        self.__last_failed_rr__ = None
        self.__last_failed_n__  = None

        self.__verified__ = False


    # --------------------------------------------------------------------------
    #
    def verify(self) -> None:

        if not self.nodes:
            return

        self.uniform = True
        node_0  = self.nodes[0]
        for node in self.nodes[1:]:

            if node.cores != node_0.cores or \
               node.gpus  != node_0.gpus  or \
               node.lfs   != node_0.lfs   or \
               node.mem   != node_0.mem:
                self.uniform = False
                break

        if self.uniform:
            self.cores_per_node = len(node_0.cores)
            self.gpus_per_node  = len(node_0.gpus)
            self.lfs_per_node   = node_0.lfs
            self.mem_per_node   = node_0.mem

        else:
            self.cores_per_node = None
            self.gpus_per_node  = None
            self.lfs_per_node   = None
            self.mem_per_node   = None

        self.__nodes_by_name__  = {node.name : node for node in self.nodes}

        self.__verified__ = True


    # --------------------------------------------------------------------------
    #
    def _assert_rr(self, rr: RankRequirements, n_slots:int) -> None:

        if not self.__verified__:
            self.verify()

        if not self.uniform:
            raise RuntimeError('verification unsupported for non-uniform nodes')

        if not rr.n_cores:
            raise ValueError('invalid rank requirements: %s' % rr)

        ranks_per_node = self.cores_per_node / rr.n_cores

        if rr.n_gpus:
            ranks_per_node = min(ranks_per_node, self.gpus_per_node / rr.n_gpus)

        if rr.lfs:
            ranks_per_node = min(ranks_per_node, self.lfs_per_node / rr.lfs)

        if rr.mem:
            ranks_per_node = min(ranks_per_node, self.mem_per_node / rr.mem)

        if ranks_per_node < 1:
            raise ValueError('invalid rank requirements: %s' % rr)

        if n_slots > len(self.nodes) * ranks_per_node:
            raise ValueError('invalid rank requirements: %s' % rr)


    # --------------------------------------------------------------------------
    #
    def find_slots(self, rr: RankRequirements, n_slots:int = 1) -> List[Slot]:

        self._assert_rr(rr, n_slots)

        if self.__last_failed_rr__:
            if self.__last_failed_rr__ >= rr and \
               self.__last_failed_n__  >= n_slots:
                return None

        slots = list()
        count = 0
        start = self.__index__

        for i in range(0, len(self.nodes)):

            idx  = (start + i) % len(self.nodes)
            node = self.nodes[idx]

            while True:
                count += 1
                slot = node.find_slot(rr)
                if not slot:
                    break

                slots.append(slot)
                if len(slots) == n_slots:
                    stop = idx
                    break

            if len(slots) == n_slots:
                break

        if len(slots) != n_slots:
            # free whatever we got
            for slot in slots:
                node = self.nodes[slot.node_index]
                node.deallocate_slot(slot)
            self.__last_failed_rr__ = rr
            self.__last_failed_n__  = n_slots
            return None

        self.__index__ = stop

        return slots


    # --------------------------------------------------------------------------
    #
    def release_slots(self, slots: List[Slot]) -> None:

        for slot in slots:

            node = self.nodes[slot.node_index]
            node.deallocate_slot(slot)

        if self.__last_failed_rr__:
            self.__index__ = min([slot.node_index for slot in slots]) - 1

        self.__last_failed_rr__ = None
        self.__last_failed_n__  = None


# ------------------------------------------------------------------------------
#
class NumaDomain(FastTypedDict):
    '''
    A `NumaDomain` defines what cores and GPUs on a node are part of the same
    momory domain.
    '''

    CORES = 'cores'
    GPUS  = 'gpus'

    _schema = {
            CORES: [int],
            GPUS : [int],
    }

    _defaults = {
            CORES: list(),
            GPUS : list(),
    }


    # --------------------------------------------------------------------------
    #
    def __init__(self, cores, gpus):

        super().__init__(cores=cores, gpus=gpus)



# ------------------------------------------------------------------------------
#
class NumaDomainMap(FastTypedDict):
    '''
    The `NumaDomainMap` is a dictionary of `NumaDomain`s which are mapped to the
    Node's NUMA domain indizes.  For example, a node with 8 cores and 2 GPUs
    could be split into two NUMA domains with 4 cores and 1 GPU each.  The
    `NumaDomainMap` would then look like this:

        {
           0: NumaDomain(cores=[0, 1, 2, 3], gpus=[0]),
           1: NumaDomain(cores=[4, 5, 6, 7], gpus=[1]),
        }
    '''

    pass


# ------------------------------------------------------------------------------
#
class NumaNode(Node):
    '''
    `NumaNode` is a specialization of `Node` which is aware of NUMA domains.
    It is used by the scheduler to allocate resources in a NUMA aware way.
    The `NumaNode` instance behaves like a regular `Node` instance if no
    `numa_domain_map` is provided.
    '''

    NUMA_DOMAINS = 'numa_domains'

    _schema = {
        NUMA_DOMAINS: {1: NumaDomain},
    }

    _defaults = {
        NUMA_DOMAINS: dict(),
    }


    # --------------------------------------------------------------------------
    #
    def as_dict(self, *args, **kwargs):

        ret = super().as_dict(*args, **kwargs)

        if ret.get('numa_domains'):
            del ret['cores']
            del ret['gpus']

        return ret


    # --------------------------------------------------------------------------
    #
    def __init__(self, from_dict: dict, numa_domain_map: dict = None) -> None:

        super().__init__(from_dict)

        if not numa_domain_map:

            # this instance behaves like a regular Node instance
            self.numa_domains = None

        else:

            # otherwise we fill the `numa_domains` map with virtual node instances,
            # one per NUMA domain
            for domain_id, domain_descr in numa_domain_map.items():

                n = Node(from_dict)
                n.name  = '%s.%s' % (self.name, domain_id)

                # the resource occupations are *shared* between the node (base
                # class) which does non-numa scheduling, and the numa domains.
                n.cores = [self.cores[i] for i in domain_descr.cores]
                n.gpus  = [self.gpus[i]  for i in domain_descr.gpus]

                # LFS and MEM are not split across NUMA domains and are always
                # allocated on node level
                n.lfs   = None
                n.mem   = None

                self.numa_domains[domain_id] = n


    # --------------------------------------------------------------------------
    #
    def find_slot(self, rr: RankRequirements) -> Optional[Slot]:

        if not rr.numa:
            return super().find_slot(rr)

        if not self.numa_domains:
            return super().find_slot(rr)

        for nd in self.numa_domains.values():

            # first find `lfs` and `mem` on the node level
            if self.lfs is not None:
                if rr.lfs and self.lfs < rr.lfs: return None

            if self.mem is not None:
                if rr.mem and self.mem < rr.mem: return None

            # then find cores and gpus on the numa domain level
            for nd in self.numa_domains.values():
                slot = nd.find_slot(rr)
                if slot:
                    return slot

# ------------------------------------------------------------------------------
#
class NumaNodeList(NodeList):


    NUMA_NODES = 'numa_nodes'

    _schema = {
        NUMA_NODES: [NumaNode],
    }


# ------------------------------------------------------------------------------





# pylint: disable=access-member-before-definition
#
__copyright__ = 'Copyright 2013-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import radical.utils as ru

from .resource_config import Slot
from .utils.misc      import FastTypedDict


# task modes
TASK_EXECUTABLE  = 'task.executable'
TASK_SERVICE     = 'task.service'
TASK_FUNCTION    = 'task.function'
TASK_FUNC        = 'task.function'
TASK_METHOD      = 'task.method'
TASK_METH        = 'task.method'
TASK_EVAL        = 'task.eval'
TASK_EXEC        = 'task.exec'
TASK_PROC        = 'task.proc'
TASK_SHELL       = 'task.shell'
RAPTOR_MASTER    = 'raptor.master'
RAPTOR_WORKER    = 'raptor.worker'
AGENT_SERVICE    = 'agent.service'

# task description attributes
UID              = 'uid'
NAME             = 'name'
MODE             = 'mode'

# mode: TASK_EXECUTABLE
EXECUTABLE       = 'executable'
ARGUMENTS        = 'arguments'

# mode: TASK_SERVICE
EXECUTABLE       = 'executable'
ARGUMENTS        = 'arguments'
INFO_PATTERN     = 'info_pattern'

# mode: TASK_METHOD
METHOD           = 'method'
ARGS             = 'args'
KWARGS           = 'kwargs'

# mode: TASK_FUNC
FUNCTION         = 'function'
ARGS             = 'args'
KWARGS           = 'kwargs'

# mode: TASK_EXEC, TASK_EVAL
CODE             = 'code'

# mode: TASK_PROC, TASK_SHELL
COMMAND          = 'command'

# mode: RAPTOR_MASTER, RAPTOR_WORKER
SCHEDULER        = 'scheduler'                # deprecated for `raptor_id`
RAPTOR_ID        = 'raptor_id'
WORKER_CLASS     = 'worker_class'             # deprecated for raptor_class
RAPTOR_CLASS     = 'raptor_class'
WORKER_FILE      = 'worker_file'              # deprecated for raptor_file
RAPTOR_FILE      = 'raptor_file'

# environment
ENVIRONMENT      = 'environment'
NAMED_ENV        = 'named_env'
SANDBOX          = 'sandbox'

# resource requirements
USE_MPI          = 'use_mpi'                  # default `True if RANKS > 1`
RANKS            = 'ranks'                    # ranks
RANKS_PER_NODE   = 'ranks_per_node'           # ranks per node
CORES_PER_RANK   = 'cores_per_rank'           # cores per rank
GPUS_PER_RANK    = 'gpus_per_rank'            # gpus per rank
THREADING_TYPE   = 'threading_type'           # OpenMP?
GPU_TYPE         = 'gpu_type'                 # CUDA / ROCm?
LFS_PER_RANK     = 'lfs_per_rank'             # disk space per rank
MEM_PER_RANK     = 'mem_per_rank'             # memory per rank
PARTITION        = 'partition'                # partition to run on

# deprecated
CPU_PROCESSES    = 'cpu_processes'            # ranks
CPU_PROCESS_TYPE = 'cpu_process_type'         # n/a
CPU_THREADS      = 'cpu_threads'              # cores per rank
CPU_THREAD_TYPE  = 'cpu_thread_type'          # OpenMP?

GPU_PROCESSES    = 'gpu_processes'            # gpus per rank
GPU_PROCESS_TYPE = 'gpu_process_type'         # CUDA?
GPU_THREADS      = 'gpu_threads'              # part of gpu?
GPU_THREAD_TYPE  = 'gpu_thread_type'          # ?

LFS_PER_PROCESS  = 'lfs_per_process'          # disk space per rank
MEM_PER_PROCESS  = 'mem_per_process'          # memory per rank

# task setup
INPUT_STAGING    = 'input_staging'
OUTPUT_STAGING   = 'output_staging'
STAGE_ON_ERROR   = 'stage_on_error'
PRIORITY         = 'priority'
PRE_LAUNCH       = 'pre_launch'
PRE_EXEC         = 'pre_exec'
PRE_EXEC_SYNC    = 'pre_exec_sync'
POST_LAUNCH      = 'post_launch'
POST_EXEC        = 'post_exec'
TIMEOUT          = 'timeout'
STARTUP_TIMEOUT  = 'startup_timeout'
CLEANUP          = 'cleanup'
PILOT            = 'pilot'
SLOTS            = 'slots'
STDOUT           = 'stdout'
STDERR           = 'stderr'
RESTARTABLE      = 'restartable'
TAGS             = 'tags'
SERVICES         = 'services'
METADATA         = 'metadata'


# ------------------------------------------------------------------------------
#
class TaskDescription(FastTypedDict):
    """Describe the requirements and properties of a Task.

    A TaskDescription object describes the requirements and properties of a
    :class:`radical.pilot.Task` and is passed as a parameter to
    :func:`radical.pilot.TaskManager.submit_tasks` to instantiate and run a new
    task.

    Attributes:
        uid (str, optional): A unique ID for the task. This attribute
            is optional, a unique ID will be assigned by RP if the field is not
            set.

        name (str, optional): A descriptive name for the task. This
            attribute can be used to map individual tasks back to application
            level workloads.

        mode (str, optional): The execution mode to be used for
            this task. Default "executable". The following modes are accepted.

            - TASK_EXECUTABLE: the task is spawned as an external executable via
              a resource specific launch method (srun, aprun, mpiexec, etc).

              - required attributes: `executable`
              - related  attributes: `arguments`

            - TASK_SERVICE: exactly like TASK_EXECUTABLE, but the task is
              handled differently by the agent.  This mode is used to start
              service tasks whose connection endpoint is made available to other
              tasks.

              NOTE: the mode `TASK_SERVICE` is only used internally and should
              not be used by application developers.

              - required attributes: `executable`
              - related  attributes: `arguments`

            - TASK_FUNCTION: the task references a python function to be called.

              - required attributes: `function`
              - related  attributes: `args`
              - related  attributes: `kwargs`

            - TASK_METHOD: the task references a raptor worker method to be
              called.

              - required attributes: `method`
              - related  attributes: `args`
              - related  attributes: `kwargs`

            - TASK_EVAL: the task is a code snippet to be evaluated.

              - required attributes: `code`

            - TASK_EXEC: the task is a code snippet to be `exec`'ed.

              - required attributes: `code`

            - TASK_SHELL: the task is a shell command line to be run.

              - required attributes: `command`

            - TASK_PROC: the task is a single core process to be executed.

              - required attributes: `executable`
              - related  attributes: `arguments`

            - TASK_RAPTOR_MASTER: the task references a raptor master to be
                  instantiated.

              - related  attributes: `raptor_file`
              - related  attributes: `raptor_class`

            - TASK_RAPTOR_WORKER: the task references a raptor worker to be
                  instantiated.

              - related  attributes: `raptor_file`
              - related  attributes: `raptor_class`

            There is a certain overlap between `TASK_EXECUTABLE`, `TASK_SHELL`
            and `TASK_PROC` modes.  As a general rule, `TASK_SHELL` and
            `TASK_PROC` should be used for short running tasks which require a
            single core and no additional resources (gpus, storage, memory).
            `TASK_EXECUTABLE` should be used for all other tasks and is in fact
            the default.  `TASK_SHELL` should only be used if the command to be
            run requires shell specific functionality (e.g., pipes, I/O
            redirection) which cannot easily be mapped to other task attributes.

            TASK_RAPTOR_MASTER and TASK_RAPTOR_WORKER are special types of tasks
            that define RAPTOR's master(s) and worker(s) components and their
            resource requirements. They are launched by the Agent on one or more
            nodes, depending on their requirements.

        executable (str): The executable to launch. The executable is
            expected to be either available via ``$PATH`` on the target
            resource, or to be an absolute path.

        arguments (list[str]): The command line arguments for the given
            `executable` (`list` of `strings`).

        info_pattern (str): A regular expression pattern to extract service
            startup information (only for tasks with mode `TASK_SERVICE`).  The
            pattern is formed like:

                'src:regex'

            where `src` is `stdout`, `stderr`, or a file path, and regex is a
            regular expression to extract the information from the respective
            source.

        code (str): The code to run.  This field is expected to contain valid
            python code which is executed when the task mode is `TASK_EXEC` or
            `TASK_EVAL`.

        function (str): The function to run.  This field is expected to contain
            a python function name which can be resolved in the scope of the
            respective RP worker implementation (see documentation there).  The
            task mode must be set to `TASK_FUNCTION`.  `args` and `kwargs` are
            passed as function parameters.

        args (list, optional): Positional arguments to be passed to the
            `function` (see above).  This field will be serialized  with
            `msgpack` and can thus contain any serializable data types.

        kwargs (dict, optional): Named arguments to be passed to the `function`
            (see above).  This field will be serialized  with `msgpack` and can
            thus contain any serializable data types.

        command (str): A shell command to be executed. This attribute is used
            for the `TASK_SHELL` mode.

        use_mpi (bool, optional): flag if the task should be provided an MPI
            communicator.  Defaults to `True` if more than 1 rank is requested
            (see `ranks`), otherwise defaults to `False`.  Set this to `True` if
            you want to enfoce an MPI communicator on single-ranked tasks.

        ranks (int, optional): The number of application processes to start
            on CPU cores. Default 1.

            For two ranks or more, an MPI communicator will be available to the
            processes.

            `ranks` replaces the deprecated attribute `cpu_processes`.  The
            attribute `cpu_process_type` was previously used to signal the need
            for an MPI communicator - that attribute is now also deprecated and
            will be ignored.

        ranks_per_node (int, optional): The number of ranks to start on each
            node.  If not set, the number of ranks per node will be determined
            by the scheduler depending on resource availability.

        cores_per_rank (int, optional): The number of cpu cores each process
            will have available to start its own threads or processes on.  By
            default, `core` refers to a physical CPU core - but if the pilot has
            been launched with SMT-settings > 1, `core` will refer to a virtual
            core or hyperthread instead (the name depends on the CPU vendor).

            `cores_per_rank` replaces the deprecated attribute `cpu_threads`.

        threading_type (str, optional): The thread type, influences startup and
            environment (`<empty>/POSIX`, `OpenMP`).

            `threading_type` replaces the deprecated attribute
            `cpu_thread_type`.

        gpus_per_rank (float, optional): The number of gpus made available to
            each rank. If gpu is shared among several ranks, then a fraction of
            gpu should be provided (e.g., 2 ranks share a GPU, then
            *gpus_per_rank=.5*).

            `gpus_per_rank` replaces the deprecated attribute `gpu_processes`.
            The attributes `gpu_threads` and `gpu_process_type` are also
            deprecated and will be ignored.

        gpu_type (str, optional): The type of GPU environment to provide to
            the ranks (`<empty>`, `CUDA`, `ROCm`).

            `gpu_type` replaces the deprecated attribute `gpu_thread_type`.

        lfs_per_rank (int, optional): Local File Storage per rank - amount of
            data (MB) required on the local file system of the node.

            `lfs_per_rank` replaces the deprecated attribute `lfs_per_process`.

        mem_per_rank (int, optional): Amount of physical memory required per
            rank.

            `mem_per_rank` replaces the deprecated attribute `mem_per_process`.

        environment (dict, optional): Environment variables to set in the
            environment before the execution (launching picked `LaunchMethod`).

        named_env (str, optional): A named virtual environment as prepared by
            the pilot. The task will remain in `AGENT_SCHEDULING` state until
            that environment gets created.

        sandbox (str, optional): This specifies the working directory of the
            task.  It will be created if it does not exist. By default, the
            sandbox has the name of the task's uid and is relative to the
            pilot's sandbox.

        stdout (str, optional): The name of the file to store stdout. If
            not set then :file:`{uid}.out` will be used.

        stderr (str, optional): The name of the file to store stderr. If
            not set then :file:`{uid}.err` will be used.

        input_staging (list, optional): The files that need to be staged before
            the execution (`list` of `staging directives`, see below).

        output_staging (list, optional): The files that need to be staged after
            the execution (`list` of `staging directives`, see below).

        stage_on_error (bool, optional): Output staging is normally skipped on
            `FAILED` or `CANCELED` tasks, but if this flag is set, staging is
            attempted either way. This may though lead to additional errors if
            the tasks did not manage to produce the expected output files to
            stage. Default False.

        priority: (int, optional): The priority of the task.  Tasks with higher
            priority will be scheduled first.  The default priority is 0.
            Note that task priorities are not guaranteed to be enforced
            strictly, under certain conditions the task may be started later
            than lower priority tasks.

        pre_launch (list, optional): Actions (shell commands) to perform
            before launching (i.e., before LaunchMethod is submitted),
            potentially on a batch node which is different from the node the
            task is placed on.

            Note that the set of shell commands given here are expected to load
            environments, check for work directories and data, etc. They are not
            expected to consume any significant amount of CPU time or other
            resources! Deviating from that rule will likely result in reduced
            overall throughput.

        post_launch (list, optional): Actions (shell commands) to perform
            after launching (i.e., after LaunchMethod is executed).

            Precautions are the same as for `pre_launch` actions.

        pre_exec (list, optional): Actions (shell commands) to perform
            before the task starts (LaunchMethod is submitted, but no actual
            task running yet). Each item could be either a string (`str`), which
            represents an action applied to all ranks, or a dictionary (`dict`),
            which represents a list of actions applied to specified ranks (key
            is a rankID and value is a list of actions to be performed for this
            rank).

            The actions/commands are executed on the respective nodes where the
            ranks are placed, and the actual rank startup will be delayed until
            all `pre_exec` commands have completed.

            Precautions are the same as for `pre_launch` actions.

            No assumption should be made as to where these commands are executed
            (although RP attempts to perform them in the task's execution
            environment).

            No assumption should be made on the specific shell environment the
            commands are executed in other than a POSIX shell environment.

            Errors in executing these commands will result in the task to enter
            `FAILED` state, and no execution of the actual workload will be
            attempted.

        pre_exec_sync (bool, optional): Flag indicates necessary to sync ranks
            execution, which enforce to delay individual rank execution, until
            all `pre_exec` actions for all ranks are completed. Default False.

        post_exec (list, optional): Actions (shell commands) to perform
            after the task finishes. The same remarks as on `pre_exec` apply,
            inclusive the point on error handling, which again will cause the
            task to fail, even if the actual execution was successful.

        restartable (bool, optional): If the task starts to execute on a pilot,
            but cannot finish because the pilot fails or is canceled, the task
            can be restarted. Default False.

        tags (dict, optional): Configuration specific tags, which
            influence task scheduling and execution (e.g., tasks co-location).

        scheduler (str, optional): deprecated in favor of `raptor_id`.

        raptor_id (str, optional): Raptor master ID this task is associated
            with.

        worker_class (str, optional): deprecated in favor of `raptor_class`
            master or worker task.

        raptor_class (str, optional): Class name to instantiate for this Raptor
            master or worker task.

        worker_file (str, optional): deprecated in favor of `raptor_class`.

        raptor_file (str, optional): Optional application supplied Python
            source file to load `raptor_class` from.

        metadata (dict, optional): User defined metadata. Default None.

        services ([str], optional): list of service names the task wants to use.
            If the services are up and running, an envvariable `RP_INFO_XXX`
            will be set where `XXX` is the uppercased service name, and the
            variable's value is the service information obtained by the agent
            during service startup.

        timeout (float, optional): Any timeout larger than 0 will result in
            the task process to be killed after the specified amount of seconds,
            starting from the launch time, if startup_timeout is not set, or
            starting from execution start time, if startup_timeout is set and
            was tracked. The task will then end up in `CANCELED` state.

        startup_timeout (float, optional): Any value larger than 0 will abort
            the task process if it hasn't started the actual execution process,
            e.g., batch system didn't provide allocated resources or the service
            startup takes too long. This option is very important for service
            tasks. The service task will end up in `FAILED` state, regular task
            will end up in `CANCELED` state.

        cleanup (bool, optional): If cleanup flag is set, the pilot will
            delete the entire task sandbox upon termination. This includes all
            generated output data in that sandbox. Output staging will be
            performed before cleanup. Note that task sandboxes are also deleted
            if the pilot's own `cleanup` flag is set. Default False.

        pilot (str, optional): Pilot `uid`. If specified, the task is
            submitted to the pilot with the given ID. If that pilot is not known
            to the TaskManager, an exception is raised.

        slots (radical.pilot.Slots, optional): information on where exactly each
            rank of the task should be placed.

        partition (int, optional): index of pilot partition to use to run that
            task.

    **Task Ranks**

    The notion of `ranks` is central to RP's `TaskDescription` class.  We here
    use the same notion as MPI, in that the number of `ranks` refers to the
    number of individual processes to be spawned by the task execution backend.
    These processes will be near-exact copies of each other: they run in the
    same workdir and the same `environment`, are defined by the same
    `executable` and `arguments`, get the same amount of resources allocated,
    etc.  Notable exceptions are:

    - Rank processes may run on different nodes;
    - rank processes can communicate via MPI;
    - each rank process obtains a unique rank ID.

    It is up to the underlying MPI implementation to determine the exact value
    of the process' rank ID.  The MPI implementation may also set a number of
    additional environment variables for each process.

    It is important to understand that only applications which make use of MPI
    should have more than one rank -- otherwise identical copies of the *same*
    application instance are launched which will compute the same results, thus
    wasting resources for all ranks but one.  Worse: I/O-routines of these
    non-MPI ranks can interfere with each other and invalidate those results.

    Also, applications with a single rank cannot make effective use of MPI---
    depending on the specific resource configuration, RP may launch those tasks
    without providing an MPI communicator.

    **Task Environment**

    RP tasks are expected to be executed in isolation, meaning that their
    runtime environment is completely independent from the environment of other
    tasks, independent from the launch mechanism used to start the task, and
    also independent from the environment of the RP stack itself.

    The task description provides several hooks to help setting up the
    environment in that context.  It is important to understand the way those
    hooks interact with respect to the environments mentioned above.

    - `pre_launch` directives are set and executed before the task is passed on
      to the task launch method.  As such, `pre_launch` usually executed on the
      node where RP's agent is running, and *not* on the tasks target node.
      Executing `pre_launch` directives for many tasks can thus negatively
      impact RP's performance (*).  Note also that `pre_launch` directives can
      in some cases interfere with the launch method.

      Use `pre_launch` directives for rare, heavy-weight operations which
      prepare the runtime environment for multiple tasks: fetch data from a
      remote source, unpack input data, create global communication channels,
      etc.
    - `pre_exec` directives are set and executed *after* the launch method
      placed the task on the compute nodes and are thus running on the target
      node.  Note that for MPI tasks, the `pre_exec` directives are executed
      once per rank.  Running large numbers of `pre_exec` directives
      concurrently can lead to system performance degradation (*), for example
      when those  directives concurrently hot the shared files system (for
      loading modules or Python virtualenvs etc).

      Use `pre_exec` directives for task environment setup such as `module
      load`, `virtualenv activate`, `export` whose effects are expected to be
      applied either to all task ranks or to specified ranks.  Avoid file
      staging operations at this point (files would be redundantly staged
      multiple times - once per rank).

    (*) The performance impact of repeated concurrent access to the system's
    shared file system can be significant and can pose a major bottleneck for
    your application.  Specifically `module load` and `virtualenv activate`
    operations and the like are heavy on file system I/O, and executing those
    for many tasks is ill advised.  Having said that: RP attempts to optimize
    those operations: if it identifies that identical `pre_exec` directives are
    shared between multiple tasks, RP will execute the directives exactly *once*
    and will cache the resulting environment settings - those cached settings
    are then applied to all other tasks with the same directives, without
    executing the directives again.

    **Staging Directives**

    The Staging Directives are specified using a dict in the following form::

          staging_directive = {
             'source'  : None, # see 'Location' below
             'target'  : None, # see 'Location' below
             'action'  : None, # See 'Action operators' below
             'flags'   : None, # See 'Flags' below
             'priority': 0     # Control ordering of actions (unused)
          }

    *Locations*

    `source` and `target` locations can be given as strings or `ru.Url`
    instances.  Strings containing `://` are converted into URLs immediately.
    Otherwise, they are considered absolute or relative paths and are then
    interpreted in the context of the client's working directory.

    .. rubric:: Special URL schemas

    - ``client://``   : relative to the client's working directory
    - ``resource://`` : relative to the RP    sandbox on the target resource
    - ``pilot://``    : relative to the pilot sandbox on the target resource
    - ``task://``     : relative to the task  sandbox on the target resource

    In all these cases, the `hostname` element of the URL is expected to be
    empty, and the path is *always* considered relative to the locations
    specified above (even though URLs usually don't have a notion of relative
    paths).

    For more details on path and sandbox handling check the documentation of
    :func:`radical.pilot.staging_directives.complete_url`.

    *Action operators*

    - rp.TRANSFER : remote file transfer from `source` to `target` URL (client)
    - rp.COPY     : local file copy, i.e., not crossing host boundaries
    - rp.MOVE     : local file move
    - rp.LINK     : local file symlink
    - rp.DOWNLOAD : fetch remote file from `source` URL to `target` URL (agent)

    *Flags*

    - rp.CREATE_PARENTS: create the directory hierarchy for targets on the fly
    - rp.RECURSIVE: if `source` is a directory, handles it recursively
    """

    _schema = {
        UID             : str         ,
        NAME            : str         ,
        MODE            : str         ,

        EXECUTABLE      : str         ,
        ARGUMENTS       : [str]       ,
        INFO_PATTERN    : str         ,
        CODE            : str         ,
        FUNCTION        : str         ,
        ARGS            : [None]      ,
        KWARGS          : {str: None} ,
        COMMAND         : str         ,

        SANDBOX         : str         ,
        ENVIRONMENT     : {str: str}  ,
        NAMED_ENV       : str         ,
        PRE_LAUNCH      : [str]       ,
        PRE_EXEC        : [None]      ,
        PRE_EXEC_SYNC   : bool        ,
        POST_LAUNCH     : [str]       ,
        POST_EXEC       : [None]      ,
        STDOUT          : str         ,
        STDERR          : str         ,
        INPUT_STAGING   : [None]      ,
        OUTPUT_STAGING  : [None]      ,
        STAGE_ON_ERROR  : bool        ,
        PRIORITY        : int         ,

        USE_MPI         : bool        ,
        RANKS           : int         ,
        RANKS_PER_NODE  : int         ,
        CORES_PER_RANK  : int         ,
        GPUS_PER_RANK   : float       ,
        THREADING_TYPE  : str         ,
        GPU_TYPE        : str         ,
        LFS_PER_RANK    : int         ,
        MEM_PER_RANK    : int         ,

        # deprecated
        CPU_PROCESSES   : int         ,  # RANKS
        CPU_PROCESS_TYPE: str         ,  # n/a
        CPU_THREADS     : int         ,  # CORES_PER_RANK
        CPU_THREAD_TYPE : str         ,  # THREADING_TYPE
        GPU_PROCESSES   : int         ,  # GPUS_PER_RANK
        GPU_PROCESS_TYPE: str         ,  # GPU_TYPE
        GPU_THREADS     : int         ,  # n/a
        GPU_THREAD_TYPE : str         ,  # n/a
        LFS_PER_PROCESS : int         ,  # LFS_PER_RANK
        MEM_PER_PROCESS : int         ,  # MEM_PER_RANK
        SCHEDULER       : str         ,  # RAPTOR_ID
        WORKER_FILE     : str         ,  # RAPTOR_FILE
        WORKER_CLASS    : str         ,  # RAPTOR_CLASS

        RESTARTABLE     : bool        ,
        TAGS            : {None: None},
        RAPTOR_ID       : str         ,
        RAPTOR_FILE     : str         ,
        RAPTOR_CLASS    : str         ,
        METADATA        : {str: None} ,
        SERVICES        : [str]       ,
        TIMEOUT         : float       ,
        STARTUP_TIMEOUT : float       ,
        CLEANUP         : bool        ,
        PILOT           : str         ,
        SLOTS           : [Slot]      ,
        PARTITION       : int         ,
    }

    _defaults = {
        UID             : ''          ,
        NAME            : ''          ,
        MODE            : TASK_EXECUTABLE,
        EXECUTABLE      : ''          ,
        ARGUMENTS       : list()      ,
        INFO_PATTERN    : ''          ,
        CODE            : ''          ,
        FUNCTION        : ''          ,
        ARGS            : list()      ,
        KWARGS          : dict()      ,
        COMMAND         : ''          ,

        SANDBOX         : ''          ,
        ENVIRONMENT     : dict()      ,
        NAMED_ENV       : ''          ,
        PRE_LAUNCH      : list()      ,
        PRE_EXEC        : list()      ,
        PRE_EXEC_SYNC   : False       ,
        POST_LAUNCH     : list()      ,
        POST_EXEC       : list()      ,
        STDOUT          : ''          ,
        STDERR          : ''          ,
        INPUT_STAGING   : list()      ,
        OUTPUT_STAGING  : list()      ,
        STAGE_ON_ERROR  : False       ,
        PRIORITY        : 0           ,

        USE_MPI         : None        ,
        RANKS           : 1           ,
        RANKS_PER_NODE  : None        ,
        CORES_PER_RANK  : 1           ,
        GPUS_PER_RANK   : 0.          ,
        THREADING_TYPE  : ''          ,
        GPU_TYPE        : ''          ,
        LFS_PER_RANK    : 0           ,
        MEM_PER_RANK    : 0           ,

        # deprecated
        CPU_PROCESSES   : 0           ,
        CPU_PROCESS_TYPE: ''          ,
        CPU_THREADS     : 0           ,
        CPU_THREAD_TYPE : ''          ,
        GPU_PROCESSES   : 0           ,
        GPU_PROCESS_TYPE: ''          ,
        GPU_THREADS     : 0           ,
        GPU_THREAD_TYPE : ''          ,
        LFS_PER_PROCESS : 0           ,
        MEM_PER_PROCESS : 0           ,
        SCHEDULER       : ''          ,
        WORKER_FILE     : ''          ,
        WORKER_CLASS    : ''          ,

        RESTARTABLE     : False       ,
        TAGS            : dict()      ,
        RAPTOR_ID       : ''          ,
        RAPTOR_FILE     : ''          ,
        RAPTOR_CLASS    : ''          ,
        METADATA        : dict()      ,
        SERVICES        : list()      ,
        TIMEOUT         : 0.0         ,
        STARTUP_TIMEOUT : 0.0         ,
        CLEANUP         : False       ,
        PILOT           : ''          ,
        SLOTS           : list()      ,
        PARTITION       : 0           ,
    }


    # --------------------------------------------------------------------------
    #
    def __init__(self, from_dict=None):

        super().__init__(from_dict=from_dict)


    # --------------------------------------------------------------------------
    #
    def _verify(self):

        if not self.get('mode'):
            self['mode'] = TASK_EXECUTABLE

        if self.mode in [TASK_EXECUTABLE, TASK_SERVICE, AGENT_SERVICE]:
            if not self.get('executable'):
                umode = self.mode.upper().replace('.', '_')
                raise ValueError("%s Task mode needs 'executable'" % umode)

        elif self.mode in [TASK_FUNC, TASK_METH]:
            if not self.get('function'):
                raise ValueError("TASK_FUNC Task mode needs 'function'")
            if self.get('named_env'):
                raise ValueError("TASK_FUNC and TASK_METH Task mode does not "
                                 "support 'named_env'")

        elif self.mode == TASK_PROC:
            if not self.get('executable'):
                raise ValueError("TASK_PROC Task mode needs 'executable'")

        elif self.mode == TASK_EVAL:
            if not self.get('code'):
                raise ValueError("TASK_EVAL Task mode needs 'code'")

        elif self.mode == TASK_EXEC:
            if not self.get('code'):
                raise ValueError("TASK_EXEC Task mode needs 'code'")

        elif self.mode == TASK_SHELL:
            if not self.get('command'):
                raise ValueError("TASK_SHELL Task mode needs 'command'")

        # backward compatibility for deprecated attributes
        if self.cpu_processes:
            self.ranks = self.cpu_processes
            self.cpu_processes = 0

        if self.cpu_threads:
            self.cores_per_rank = self.cpu_threads
            self.cpu_threads = 0

        if self.cpu_thread_type:
            self.threading_type = self.cpu_thread_type
            self.cpu_thread_type = None

        if self.gpu_processes:
            self.gpus_per_rank = float(self.gpu_processes)
            self.gpu_processes = 0

        if self.gpu_process_type:
            self.gpu_type = self.gpu_process_type
            self.gpu_process_type = None

        if self.lfs_per_process:
            self.lfs_per_rank = self.lfs_per_process
            self.lfs_per_process = 0

        if self.mem_per_process:
            self.mem_per_rank = self.mem_per_process
            self.mem_per_process = 0

        if self.scheduler:
            self.raptor_id = self.scheduler
            self.scheduler = ''

        if self.worker_file:
            self.raptor_file = self.worker_file
            self.worker_file = ''

        if self.worker_class:
            self.raptor_class = self.worker_class
            self.raptor_class = ''

        if self.use_mpi is None:
            self.use_mpi = bool(self.ranks - 1)

        # deprecated and ignored
        if self.cpu_process_type: pass
        if self.gpu_threads     : pass
        if self.gpu_thread_type : pass

      # if self.mode in [TASK_SHELL, TASK_PROC]:
      #
      #     if self.get('cpu_processes', 1) * self.get('cpu_threads', 1) > 1:
      #         raise ValueError("TASK_SHELL and TASK_PROC Tasks must be single core")
      #
      #     if self.get('gpu_processes', 0) > 0:
      #         raise ValueError("TASK_SHELL and TASK_PROC Tasks canont use GPUs")


# ------------------------------------------------------------------------------




__copyright__ = 'Copyright 2016-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class MPIRun(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, lm_cfg, rm_info, log, prof):

        self._mpt    : bool  = False
        self._rsh    : bool  = False
        self._ccmrun : str   = ''
        self._dplace : str   = ''
        self._omplace: str   = ''
        self._command: str   = ''

        LaunchMethod.__init__(self, name, lm_cfg, rm_info, log, prof)


    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, env, env_sh):
        '''
        The RP launch methods are used in different components: the agent will
        use them to spawn sub-agents and components, the resource manager may
        use them to spawn launcher services, and, of course, different execution
        components (including Raptor and other task overlays) can use them to
        launch tasks.

        The first use (likely in `agent_0`) will call this initializer to
        inspect LM properties.  Later uses will be able to use the information
        gathered and should re-initialize via `_init_from_info()`, using the
        info dict returned here.
        '''

        lm_info = {
            'env'    : env,
            'env_sh' : env_sh,
            'command': ru.which([
                'mpirun',             # general case
                'mpirun_rsh',         # Gordon (SDSC)
                'mpirun_mpt',         # Cheyenne (NCAR)
                'mpirun-mpich-mp',    # Mac OSX mpich
                'mpirun-openmpi-mp',  # Mac OSX openmpi
            ]),
            'mpt'    : False,
            'rsh'    : False,
            'ccmrun' : '',
            'dplace' : '',
            'omplace': ''
        }

        if not lm_info['command']:
            raise ValueError('mpirun not found - cannot start MPI tasks')

        if '_mpt' in self.name.lower():
            lm_info['mpt'] = True

        if '_rsh' in self.name.lower():
            lm_info['rsh'] = True

        # do we need ccmrun or dplace?
        if '_ccmrun' in self.name.lower():
            lm_info['ccmrun'] = ru.which('ccmrun')
            assert lm_info['ccmrun']

        if '_dplace' in self.name.lower():
            lm_info['dplace'] = ru.which('dplace')
            assert lm_info['dplace']

        # cheyenne always needs mpt and omplace
        if 'cheyenne' in ru.get_hostname():
            lm_info['omplace'] = ru.which('omplace')
            lm_info['mpt']     = True

        try:
            mpi_version, mpi_flavor = self._get_mpi_info(lm_info['command'])
            lm_info['mpi_version']  = mpi_version
            lm_info['mpi_flavor']   = mpi_flavor
            self._log.debug('got mpi version: %s (%s)', mpi_version, mpi_flavor)
        except:
            self._log.exception('failed to inspect MPI version and flavor')

        return lm_info


    # --------------------------------------------------------------------------
    #
    def _init_from_info(self, lm_info):

        self._env         = lm_info['env']
        self._env_sh      = lm_info['env_sh']
        self._command     = lm_info['command']

        assert self._command

        self._mpt         = lm_info['mpt']
        self._rsh         = lm_info['rsh']
        self._ccmrun      = lm_info['ccmrun']
        self._dplace      = lm_info['dplace']
        self._omplace     = lm_info['omplace']

        self._mpi_version = lm_info['mpi_version']
        self._mpi_flavor  = lm_info['mpi_flavor']


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        pass


    # --------------------------------------------------------------------------
    #
    def can_launch(self, task):

        if not task['description']['executable']:
            return False, 'no executable'

        return True, ''


    # --------------------------------------------------------------------------
    #
    def get_launcher_env(self):

        lm_env_cmds = ['. $RP_PILOT_SANDBOX/%s' % self._env_sh]

        # Cheyenne is the only machine that requires mpirun_mpt.
        # We then have to set MPI_SHEPHERD=true
        if self._mpt:
            lm_env_cmds.append('export MPI_SHEPHERD=true')

        return lm_env_cmds


    # --------------------------------------------------------------------------
    #
    def get_launch_cmds(self, task, exec_path):

        slots      = task['slots']
        uid        = task['uid']
        td         = task['description']
        sandbox    = task['task_sandbox_path']
        task_cores = td.get('cores_per_rank', 1)
        task_gpus  = td.get('gpus_per_rank', 0.)

        if '_dplace' in self.name and task_cores > 1:
            # dplace pinning would disallow threads to map to other cores
            raise ValueError('dplace can not place threads [%d]' % task_cores)

        # Extract all the hosts from the slots
        host_list = list()
        core_list = list()
        save_list = list()

        for slot in slots:

            host_list.append(slot['node_name'])
            core_list.append(slot['cores'][0])
            # FIXME: inform this proc about the GPU to be used

            if '_dplace' in self.name and save_list:
                assert (save_list == core_list), 'inhomog. core sets (dplace)'

            else:
                save_list = core_list

        if '_dplace' in self.name:
            self._dplace += ' -c '
            self._dplace += ','.join(core_list)


        # If we have a task with many cores, we will create a hostfile and pass
        # that as an argument instead of the individual hosts
        hosts_string     = ''
        mpt_hosts_string = ''
        if len(host_list) > 42:

            # Create a hostfile from the list of hosts
            hostfile = ru.create_hostfile(sandbox, uid, host_list,
                                          impaired=True)
            if self._mpt: hosts_string = '-file %s'     % hostfile
            else        : hosts_string = '-hostfile %s' % hostfile

        else:
            # Construct the hosts_string ('h1,h2,..,hN')
            if self._mpt: mpt_hosts_string = '%s'       % ','.join(host_list)
            else        : hosts_string     = '-host %s' % ','.join(host_list)

        # -np:  usually len(host_list), meaning N processes over N hosts, but
        # for Cheyenne (mpt) the specification of -host lands N processes on
        # EACH host, where N is specified as arg to -np
        if self._mpt: np = 1
        else        : np = len(host_list)

        options = ''
        if task_gpus and self._mpi_flavor == self.MPI_FLAVOR_SPECTRUM:
            # https://on-demand.gputechconf.com/gtc/2018/presentation/
            # s8314-multi-gpu-programming-with-mpi.pdf - s.18
            options = '-gpu '

        options += '-np %d' % np

        cmd = '%s %s %s %s %s %s %s %s' % \
            (self._ccmrun, self._command, mpt_hosts_string, options,
             self._dplace, self._omplace, hosts_string, exec_path)

        return cmd.strip()


    # --------------------------------------------------------------------------
    #
    def get_rank_cmd(self):

        # FIXME: we know the MPI flavor, so make this less guesswork

        ret  = 'test -z "$MPI_RANK"     || export RP_RANK=$MPI_RANK\n'
        ret += 'test -z "$PMIX_RANK"    || export RP_RANK=$PMIX_RANK\n'

        if self._mpt:
            ret += 'test -z "$MPT_MPI_RANK" || export RP_RANK=$MPT_MPI_RANK\n'

        if self._mpi_flavor == self.MPI_FLAVOR_HYDRA:
            ret += 'test -z "$PMI_ID"       || export RP_RANK=$PMI_ID\n'
            ret += 'test -z "$PMI_RANK"     || export RP_RANK=$PMI_RANK\n'
        elif self._mpi_flavor == self.MPI_FLAVOR_PALS:
            ret += 'test -z "$PALS_RANKID"  || export RP_RANK=$PALS_RANKID\n'

        return ret


    # --------------------------------------------------------------------------
    #
    def get_env_blacklist(self):
        '''
        extend the base blacklist by SLURM specific environment variables
        '''

        ret = super().get_env_blacklist()

        ret.extend([
                    'OMPI_*',
                    'PMIX_*',
                   ])

        return ret


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import os
import sys
import time
import socket

import radical.pilot as rp


n_nodes = 1

# ------------------------------------------------------------------------------
#
def app():

    session = rp.Session()
    try:
        pmgr  = rp.PilotManager(session=session)
        tmgr  = rp.TaskManager(session=session)
        pdesc = rp.PilotDescription({'resource': 'local.localhost',
                                     'runtime' : 15,
                                     'nodes'   : n_nodes,
                                     'services': list()})

        for idx in range(n_nodes):
            pdesc['services'].append(
                rp.TaskDescription({
                    'uid'           : 'test_service_%02d' % idx,
                    'executable'     : __file__,
                    'arguments'      : ['service', 1024 + 1 + idx],
                  # 'stdout'         : '/tmp/watcher_%d.out' % idx,
                    'ranks'          : 1,
                    'cores_per_rank' : 1,
                    'named_env'      : 'rp',
                })
            )
        pilot = pmgr.submit_pilots(pdesc)
        tmgr.add_pilots(pilot)

        td = rp.TaskDescription({'executable': '/bin/date',
                                 'services'  : ['test_service_00',
                                                'test_service_01']})
        task = tmgr.submit_tasks(td)
        tmgr.wait_tasks(task.uid)
        print(' * %s: %s' % (task.uid, task.stdout))

      # pilot.register_service('ext_service', 'tcp://localhost:12345')
      #
      # td = rp.TaskDescription({'executable': '/bin/date',
      #                          'services'  : ['ext_service']})
      # task = tmgr.submit_tasks(td)
      # tmgr.wait_tasks(task.uid)

        # ---------------------------------------------------------------------
        # start SERVICE instance and wait for startup info
        sd = rp.TaskDescription(
             {'uid'          : 'my_service',
              'mode'         : rp.TASK_SERVICE,
              'executable'   : __file__,
              'arguments'    : ['service', 1024],
              'info_pattern' : 'stdout:.*port: ([0-9]*).*',
              'timeout'      : 10,  # startup timeout
              'named_env'    : 'rp',
              })
        service = tmgr.submit_tasks(sd)

        # collect the endpoint port
        info = service.wait_info()
        print(' * %s: %s' % (service.uid, info))


        # ---------------------------------------------------------------------
        # run the clients
        cds = list()
        for i in range(3):
            cd = rp.TaskDescription({'services'   : [service.uid],
                                     'executable' : __file__,
                                     'arguments'  : ['client']})
            cds.append(cd)

        clients = tmgr.submit_tasks(cds)
        tmgr.wait_tasks(uids=[client.uid for client in clients])

        for client in clients:
            print(' * %s: %s' % (client.uid, client.stdout))

    finally:
        session.close(download=True)


# ------------------------------------------------------------------------------
#
def service(port):

    port = int(port)

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.settimeout(1.0)
    sock.bind(('localhost', port))

    sys.stdout.write('port: %d\n' % port)
    sys.stdout.write('pid : %d\n' % os.getpid())
    sys.stdout.flush()

    sock.listen(1)
    start = time.time()
    while True:

        print('%.2f: service loop' % time.time())

        try:
            client, _ = sock.accept()
            client.send(client.recv(port))
            client.close()
            print('%.2f: service served' % time.time())

        except socket.timeout:
            # run for 60 seconds, at most
            if time.time() - start > 3600:
                break


# ------------------------------------------------------------------------------
#
def client():

    port = int(os.environ['RP_INFO_MY_SERVICE'])

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(('localhost', port))
    sock.send(b'foo')
    print(str(sock.recv(port)))
    sock.close()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    mode = sys.argv[1] if len(sys.argv) > 1 else 'app'
    port = sys.argv[2] if len(sys.argv) > 2 else 1024

    if   mode == 'app'    : app()
    elif mode == 'service': service(port)
    elif mode == 'client' : client()
    else: raise ValueError('unknown mode %s' % mode)


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2020-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import collections
import math
import os
import signal
import time

from typing import Dict, Union

import subprocess    as mp
import threading     as mt

import radical.utils as ru

from .base import LaunchMethod

PTL_MAX_MSG_SIZE = 1024 * 1024 * 1024 * 1

DVM_URI_FILE_TPL   = '%(base_path)s/prrte.%(dvm_id)04d.uri'
DVM_HOSTS_FILE_TPL = '%(base_path)s/prrte.%(dvm_id)04d.hosts'


# ------------------------------------------------------------------------------
#
class PRTE(LaunchMethod):
    """
    PMIx Reference Run Time Environment v2
    (https://github.com/openpmix/prrte)
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, lm_cfg, rm_info, log, prof):

        self._verbose = bool(os.environ.get('RADICAL_PILOT_PRUN_VERBOSE'))

        LaunchMethod.__init__(self, name, lm_cfg, rm_info, log, prof)

    # --------------------------------------------------------------------------
    #
    def __del__(self):

        self._terminate()

    # --------------------------------------------------------------------------
    #
    def _configure(self):

        prte_cmd = ru.which('prte')
        if not prte_cmd:
            raise Exception('prte command not found')

        prte_prefix = os.environ.get('UMS_OMPIX_PRRTE_DIR')

        prte_info = {}  # get OpenRTE/PRTE version
        out = ru.sh_callout('prte_info | grep "RTE"', shell=True)[0]
        for line in out.split('\n'):

            line = line.strip()

            if not prte_info.get('name'):
                if   'Open RTE' in line: prte_info['name'] = 'Open RTE'
                elif 'PRTE'     in line: prte_info['name'] = 'PRTE'

            if 'RTE:' in line:
                prte_info['version'] = line.split(':')[1].strip()
            elif 'RTE repo revision:' in line:
                prte_info['version_detail'] = line.split(':')[1].strip()

        if prte_info.get('name'):
            self._log.info('version of %s: %s [%s]',
                           prte_info['name'],
                           prte_info.get('version'),
                           prte_info.get('version_detail'))
        else:
            self._log.info('version of OpenRTE/PRTE not determined')

        # get `dvm_count` from resource config (LM config section)
        # FIXME: another option is to introduce `nodes_per_dvm` in LM config
        #        dvm_count = math.ceil(len(rm_info.node_list) / nodes_per_dvm)
        dvm_count     = self._lm_cfg.get('dvm_count') or 1
        nodes_per_dvm = math.ceil(len(self._rm_info.node_list) / dvm_count)

        # additional info to form prrte related files (uri-/hosts-file)
        dvm_file_info : Dict[str, Union[str,int]] = {'base_path': os.getcwd()}

        # ----------------------------------------------------------------------
        def _watch_dvm(dvm_process, dvm_id, dvm_ready_flag):

            self._log.info('prte-%s watcher started', dvm_id)

            retval = dvm_process.poll()
            while retval is None:
                line = dvm_process.stdout.readline().strip()
                if line:
                    self._log.debug('prte-%s output: %s', dvm_id, line)

                    # final confirmation that DVM has started successfully
                    if not dvm_ready_flag.is_set() and 'DVM ready' in str(line):
                        dvm_ready_flag.set()
                        self._log.info('prte-%s is ready', dvm_id)
                        self._prof.prof(event='dvm_ready',
                                        uid=self._lm_cfg['pid'],
                                        msg='dvm_id=%s' % dvm_id)

                else:
                    time.sleep(1.)

            if retval != 0:
                # send a kill signal to the main thread.
                # We know that Python and threading are likely not to play well
                # with signals - but this is an exceptional case, and not part
                # of the standard termination sequence. If the signal is
                # swallowed, the next prun-call will trigger termination anyway.
                os.kill(os.getpid(), signal.SIGKILL)
                raise RuntimeError('PRTE DVM died (dvm_id=%s)' % dvm_id)

            self._log.info('prte-%s watcher stopped', dvm_id)

        # ----------------------------------------------------------------------
        def _start_dvm(dvm_id, dvm_size, dvm_ready_flag):

            file_info = dvm_file_info
            file_info['dvm_id'] = dvm_id

            prte  = '%s' % prte_cmd
            if prte_prefix: prte += ' --prefix %s' % prte_prefix
            prte += ' --report-uri %s' % DVM_URI_FILE_TPL   % file_info
            prte += ' --hostfile %s'   % DVM_HOSTS_FILE_TPL % file_info

            # large tasks imply large message sizes, we need to account for that
            # FIXME: need to derive the message size from DVM size - smaller
            #        DVMs will never need large messages, as they can't run
            #        large tasks
            prte += ' --pmixmca ptl_base_max_msg_size %d'  % PTL_MAX_MSG_SIZE
            prte += ' --prtemca routed_radix %d'           % dvm_size

            # ensure 1 ssh per dvm (all daemons are directly connected to prte)
            prte += ' --prtemca plm_rsh_num_concurrent %d' % dvm_size

            # ssh connection limit (avoid 64 node limit)
            #   --prtemca plm_rsh_no_tree_spawn 1'

            # select a set of hardware threads that should be used for
            # processing the following option should set hwthread IDs:
            #   --prtemca hwloc_default_cpu_list "0-83,88-171"

            # Use (g)stdbuf to disable buffering.  We need this to get the
            # "DVM ready" message to ensure DVM startup completion
            #
            # The command seems to be generally available on Cray's,
            # if not, we can code some home-cooked pty stuff (TODO)
            stdbuf_cmd = ru.which(['stdbuf', 'gstdbuf'])
            if not stdbuf_cmd:
                raise Exception('(g)stdbuf command not found')
            stdbuf_arg = '-oL'

            # base command = (g)stdbuf <args> + prte <args>
            cmd = '%s %s %s ' % (stdbuf_cmd, stdbuf_arg, prte)

            # additional (debug) arguments to prte
            if self._verbose:
                cmd += ' '.join(['--prtemca plm_base_verbose 5'])

            cmd = cmd.strip()

            self._log.info('start prte-%s on %d nodes [%s]',
                           dvm_id, dvm_size, cmd)
            self._prof.prof(event='dvm_start',
                            uid=self._lm_cfg['pid'],
                            msg='dvm_id=%s' % dvm_id)

            dvm_process = mp.Popen(cmd.split(),
                                   stdout=mp.PIPE,
                                   stderr=mp.STDOUT)
            dvm_watcher = mt.Thread(target=_watch_dvm,
                                    args=(dvm_process, dvm_id, dvm_ready_flag),
                                    daemon=True)
            dvm_watcher.start()

            dvm_uri = None
            for _ in range(30):
                time.sleep(.5)

                try:
                    with ru.ru_open(DVM_URI_FILE_TPL % file_info, 'r') as fin:
                        for line in fin.readlines():
                            if '://' in line:
                                dvm_uri = line.strip()
                                break
                except Exception as e:
                    self._log.debug('DVM URI file missing: %s...', str(e)[:24])

                if dvm_uri:
                    break

            if not dvm_uri:
                raise Exception('DVM URI not found')

            self._log.info('prte-%s startup successful: [%s]', dvm_id, dvm_uri)
            self._prof.prof(event='dvm_uri',
                            uid=self._lm_cfg['pid'],
                            msg='dvm_id=%s' % dvm_id)

            return dvm_uri

        # ----------------------------------------------------------------------

        # format: {<id>: {'nodes': [...], 'dvm_uri': <uri>}}
        dvm_list = {}

        # go through every dvm instance
        for _dvm_id in range(dvm_count):

            node_list = self._rm_info.node_list[_dvm_id      * nodes_per_dvm:
                                               (_dvm_id + 1) * nodes_per_dvm]
            dvm_file_info['dvm_id'] = _dvm_id
            # write hosts file
            with ru.ru_open(DVM_HOSTS_FILE_TPL % dvm_file_info, 'w') as fout:
                for node in node_list:
                    fout.write('%s slots=%d\n' % (node['name'],
                                                  self._rm_info.cores_per_node))

            _dvm_size  = len(node_list)
            _dvm_ready = mt.Event()
            _dvm_uri   = _start_dvm(_dvm_id, _dvm_size, _dvm_ready)

            dvm_list[_dvm_id] = {
                'nodes'  : [node['index'] for node in node_list],
                'dvm_uri': _dvm_uri}

            # extra time to confirm that "DVM ready" was just delayed
            if _dvm_ready.wait(timeout=15.):
                continue

            self._log.info('prte-%s to be re-started', _dvm_id)

            # terminate DVM (if it is still running)
            try   : ru.sh_callout('pterm --dvm-uri "%s"' % _dvm_uri)
            except: pass

            # re-start DVM
            dvm_list[_dvm_id]['dvm_uri'] = \
                _start_dvm(_dvm_id, _dvm_size, _dvm_ready)

            # FIXME: with the current approach there is only one attempt to
            #        restart DVM(s). If a failure during the start process
            #        will keep appears then need to consider re-assignment
            #        of nodes from failed DVM(s) + add time for re-started
            #        DVM to stabilize: `time.sleep(10.)`

        lm_details = {'dvm_list'    : dvm_list,
                      'version_info': prte_info}

        return lm_details


    # --------------------------------------------------------------------------
    #
    def get_partitions(self):

        partitions = dict()

        for k, v in self._details['dvm_list'].items():
            partitions[k] = v['nodes']

        return partitions


    # --------------------------------------------------------------------------
    #
    def _terminate(self):
        """
        This hook is symmetric to the hook `rm_config_hook`, it is called
        during shutdown sequence, for the sake of freeing allocated resources.
        """
        cmd = ru.which('pterm')
        if not cmd:
            return

        for p_id, p_data in self._details.get('dvm_list', {}).items():
            dvm_uri = p_data['dvm_uri']
            try:
                self._log.info('terminating prte-%d (%s)', p_id, dvm_uri)
                _, err, _ = ru.sh_callout('%s --dvm-uri "%s"' % (cmd, dvm_uri))
                self._log.debug('termination state: %s', err.strip('\n') or '-')
                self._prof.prof(event='dvm_stop', uid=self._lm_cfg['pid'],
                                msg='dvm_id=%d' % p_id)

            except Exception as e:
                # use the same event name as for runtime failures - those are
                # not distinguishable at the moment from termination failures
                self._log.debug('prte-%d termination failed (%s)', p_id, dvm_uri)
                self._prof.prof(event='dvm_fail', uid=self._lm_cfg['pid'],
                                msg='dvm_id=%s | error: %s' % (p_id, e))

    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, env, env_sh):

        lm_info = {'env'    : env,
                   'env_sh' : env_sh,
                   'command': ru.which('prun'),
                   'details': self._configure()}

        return lm_info

    # --------------------------------------------------------------------------
    #
    def _init_from_info(self, lm_info):

        self._env     = lm_info['env']
        self._env_sh  = lm_info['env_sh']
        self._command = lm_info['command']
        self._details = lm_info['details']

        assert self._command

    # --------------------------------------------------------------------------
    #
    def finalize(self):

        self._terminate()

    # --------------------------------------------------------------------------
    #
    def can_launch(self, task):

        if not task['description']['executable']:
            return False, 'no executable'

        return True, ''


    # --------------------------------------------------------------------------
    #
    def get_launcher_env(self):

        return ['. $RP_PILOT_SANDBOX/%s' % self._env_sh]

    # --------------------------------------------------------------------------
    #
    def get_launch_cmds(self, task, exec_path):

        # latest implementation of PRRTE is able to handle bulk submissions
        # of prun-commands, but in case errors will come out then the delay
        # should be set: `time.sleep(.1)`

        slots     = task['slots']
        td        = task['description']
        partition = task['partition']

        n_procs   = td['ranks']
        n_threads = td['cores_per_rank']

        if not self._details.get('dvm_list'):
            raise RuntimeError('details with dvm_list not set (%s)' % self.name)

        dvm_list  = self._details['dvm_list']
        dvm_id    = partition
        dvm_uri   = '--dvm-uri "%s"' % dvm_list[dvm_id]['dvm_uri']

        flags  = '--np %d '                                   % n_procs
        flags += '--map-by node:HWTCPUS:PE=%d:OVERSUBSCRIBE ' % n_threads
        flags += '--bind-to hwthread:overload-allowed'

        if self._verbose:
            flags += ':REPORT'

        if not slots:
            # this task is unscheduled - we leave it to PRRTE/PMI-X
            # to correctly place the task
            pass
        else:
            ranks = collections.defaultdict(int)
            for slot in slots:
                ranks[slot['node_name']] += 1
            flags += ' --host ' + ','.join(['%s:%s' % x for x in ranks.items()])

        flags += ' --pmixmca ptl_base_max_msg_size %d' % PTL_MAX_MSG_SIZE
        flags += ' --verbose'  # needed to get prte profile events

        cmd = '%s %s %s %s' % (self._command, dvm_uri, flags, exec_path)

        return cmd.strip()

    # --------------------------------------------------------------------------
    #
    def get_rank_cmd(self):

        return 'export RP_RANK=$PMIX_RANK\n'


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import os
import copy
import time
import random

from collections import defaultdict

import threading as mt

import radical.pilot as rp
import radical.utils as ru


# ------------------------------------------------------------------------------
#
class Pipeline(object):

    # the pipeline renders this structure:
    #
    #  Simulation_1
    #       |
    #       |
    #       |
    #  Policies_1
    #       |    \
    #       |     \
    #       |      \
    #  Prelim_1   Training_1
    #       |       |
    #       |       |
    #       |       |
    #  Simulation_2 |
    #       |       |
    #       |       |
    #       |       |
    #  Policies_2   |
    #       |    \  |
    #       |     \ |
    #       |      \|
    #  Prelim_2   Training_2
    #       |       |
    #       |       |
    #       |       |
    #  Simulation_3 |
    #       |       |
    #       |       |
    #       |       |
    #  Policies_3   |
    #       |    \  |
    #       |     \ |
    #       |      \|
    #  Prelim_3   Training_3
    #       |       |
    #      ...     ...

    # define task types
    TASK_SIM     = 'simulation'
    TASK_POLICY  = 'policy'
    TASK_PRELIM  = 'preliminary'
    TASK_TRAIN   = 'training'

    TASK_TYPES   = [TASK_SIM, TASK_POLICY, TASK_PRELIM, TASK_TRAIN]

    task_sim     = rp.TaskDescription({'executable': 'sleep',
                                       'metadata'  : {'type': TASK_SIM,
                                                      'iter': None}})
    task_policy  = rp.TaskDescription({'executable': 'sleep',
                                       'metadata'  : {'type': TASK_POLICY,
                                                      'iter': None}})
    task_prelim  = rp.TaskDescription({'executable': 'sleep',
                                       'metadata'  : {'type': TASK_PRELIM,
                                                      'iter': None}})
    task_train   = rp.TaskDescription({'executable': 'sleep',
                                       'metadata'  : {'type': TASK_TRAIN,
                                                      'iter': None}})

    # condition indexes -- see `self._cond_train`
    COND_PRELIM = 0  # a dependent preliminary task has completed
    COND_TRAIN  = 1  # a dependent training task has completed
    COND_ACTED  = 2  # the condition has been acted upon


    # --------------------------------------------------------------------------
    #
    def __init__(self, tmgr):

        self._tmgr = tmgr

        self._uid       = ru.generate_id('p.%(item_counter)1d', ru.ID_CUSTOM)
        self._iteration = None  # count generation of TASK_SIM instances

        # control flow table
        self._protocol = {self.TASK_SIM   : self._control_sim,
                          self.TASK_POLICY: self._control_policy,
                          self.TASK_PRELIM: self._control_prelim,
                          self.TASK_TRAIN : self._control_train}


        # TASK_TRAIN's have two dependencies: the completed TASK_PRELIM of the
        # same iteration, and the completed TASK_TRAIN of the *previous*
        # iteration (for any iteration >1).  We start the TASK_TRAIN on
        # whichever of those two preconditions is met last.  self._cond_train
        # will keep track of the preconditions and also of actions on met
        # conditions
        self._cond_lock  = mt.Lock()
        self._cond_train = defaultdict(lambda: [False, False, False])

        # for iteration 0, we mark the second precondition as resolved: there is
        # no previous iteration to wait on
        self._cond_train[0][self.COND_TRAIN] = True

        # bookkeeping: what task type in what iteration is active right now?
        self._stats = {self.TASK_SIM   : None,
                       self.TASK_POLICY: None,
                       self.TASK_PRELIM: None,
                       self.TASK_TRAIN : None}


    # --------------------------------------------------------------------------
    #
    def dump(self, title=False, header=False):
        '''
        dump a representation of current task set in a string
        '''

        if title:
            out = ' | %-33s | ' % self._uid
            return out

        if header:
            out = ' | '
            for ttype in ['SIM', 'PRELIM', 'POLICY', 'TRAIN']:

                out += '% 6s | ' % ttype

            return out

        out = ' | '
        for ttype in [self.TASK_SIM,    self.TASK_PRELIM,
                      self.TASK_POLICY, self.TASK_TRAIN]:

            iteration = self._stats[ttype]
            if iteration is None:
                out += '%6s | ' % ''
            else:
                out += '%6d | ' % self._stats[ttype]

        return out


    # --------------------------------------------------------------------------
    #
    def run(self):
        '''
        submit initial set of MD similation tasks
        '''

        self._iteration = 0

        # run initial MD_SIM task
        self._submit_task(self.TASK_SIM, self._iteration)


    # --------------------------------------------------------------------------
    #
    def stop(self):

        raise RuntimeError('stop %s' % self._uid)


    # --------------------------------------------------------------------------
    #
    def _get_tinfo(self, task):
        '''
        get task type and iteration from task metadata
        '''

        ttype =     task.metadata['type']
        titer = int(task.metadata['iter'])

        assert ttype in self.TASK_TYPES, 'unknown task type: %s' % ttype
        return ttype, titer


    # --------------------------------------------------------------------------
    #
    def _submit_task(self, ttype, iteration):
        '''
        submit new task of specified type in given iteration
        '''

        if   ttype == self.TASK_SIM   : td = copy.deepcopy(self.task_sim)
        elif ttype == self.TASK_POLICY: td = copy.deepcopy(self.task_policy)
        elif ttype == self.TASK_PRELIM: td = copy.deepcopy(self.task_prelim)
        elif ttype == self.TASK_TRAIN : td = copy.deepcopy(self.task_train)
        else: raise ValueError('no such task type %s' % ttype)

        td['uid']              = '%s.%s.%03d' % (self._uid, ttype, iteration)
        td['arguments']        = random.randint(0, 10)
        td['metadata']['iter'] = iteration

        task = self._tmgr.submit_tasks([td])[0]
        task.register_callback(self._state_cb)

        self._stats[ttype] = iteration
        self.dump()


    # --------------------------------------------------------------------------
    #
    def _state_cb(self, task, state):
        '''
        act on task state changes according to our protocol
        '''

        try:
            # this cb will react on task state changes.  Specifically it will watch
            # out for task completion notification and react on them, depending on
            # the task type.

            # we only handle final states
            if state not in rp.FINAL:
                return

            if state == rp.CANCELED:
                print('task %s cancelled: stop' % task.uid)
                self.stop()

            if state == rp.FAILED:
                print('task %s failed: %s' % (task.uid, task.stderr))
                self.stop()

            assert state == rp.DONE


            # control flow depends on ttype
            ttype, titer = self._get_tinfo(task)
            self._stats[ttype]  = None
            self.dump()

            action = self._protocol[ttype]
            action(task, titer)

        except Exception as e:
            print('\n\n---------\nexception caught: %s\n\n' % repr(e))
            self.stop()


    # --------------------------------------------------------------------------
    #
    def _control_sim(self, task, iteration):
        '''
        when a simulation task completes, a preliminary training task is started
        in the same iteration.
        '''

        self._submit_task(self.TASK_PRELIM, iteration)


    # --------------------------------------------------------------------------
    #
    def _control_prelim(self, task, iteration):
        '''
        when a preliminary training task completes, two things happen:
          - a learning policy task is started in the same iteration
          - a training task is started in the same iteration

        However, the training task will *only* be started if the training task
        of the *previous* iteration has also completed.
        '''

        # always start the learning policy task
        self._submit_task(self.TASK_POLICY, iteration)

        with self._cond_lock:

            # mark condition as resolved
            self._cond_train[iteration][self.COND_PRELIM] = True

            # if all conditions are met, start training task and mark as started
            if self._cond_train[iteration] == [True, True, False]:
                self._submit_task(self.TASK_TRAIN, iteration)
                self._cond_train[iteration][self.COND_ACTED] = True


    # --------------------------------------------------------------------------
    #
    def _control_policy(self, task, iteration):
        '''
        when a learning policy task completes, start a new simulation task of
        the *next* iteration
        '''

        self._submit_task(self.TASK_SIM, iteration + 1)


    # --------------------------------------------------------------------------
    #
    def _control_train(self, task, iteration):
        '''
        when a training task completes, mark the training condition of the
        *next* iteration as fullfilled.  If the next iteration has both
        conditions fullfilled for it's training task, then start it also.
        '''

        with self._cond_lock:

            # mark condition as resolved
            self._cond_train[iteration + 1][self.COND_TRAIN] = True

            # if all conditions are met, start training task and mark as started
            if self._cond_train[iteration + 1] == [True, True, False]:
                self._submit_task(self.TASK_TRAIN, iteration + 1)
                self._cond_train[iteration + 1][self.COND_ACTED] = True


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # silence RP reporter
    os.environ['RADICAL_REPORT'] = 'false'

    with rp.Session() as session:

        # RP setup
        pmgr    = rp.PilotManager(session=session)
        tmgr    = rp.TaskManager(session=session)

        pdesc = rp.PilotDescription({'resource': 'local.localhost',
                                     'runtime' : 30,
                                     'cores'   : 10})
        pilot = pmgr.submit_pilots(pdesc)

        tmgr.add_pilots(pilot)

        # create pipelines
        pipeline_1 = Pipeline(tmgr)
        pipeline_2 = Pipeline(tmgr)

        out  = pipeline_1.dump(title=True)
        out += pipeline_2.dump(title=True)
        print(out)

        # start pipelines
        pipeline_1.run()
        pipeline_2.run()


        # run them forever (well, until the pilot dies...
        i = 0
        while True:

            if not i % 30:
                out  = pipeline_1.dump(header=True)
                out += pipeline_2.dump(header=True)
                print(out)

            out  = pipeline_1.dump()
            out += pipeline_2.dump()
            print(out)

            time.sleep(1)
            i += 1


# ------------------------------------------------------------------------------

